// --*- c++ -*---
#ifndef __SYCL_DATA_LESS_PROP
#define __SYCL_DATA_LESS_PROP(NS_QUALIFIER, PROP_NAME, ENUM_VAL)
#endif
#ifndef __SYCL_MANUALLY_DEFINED_PROP
#define __SYCL_MANUALLY_DEFINED_PROP(NS_QUALIFIER, PROP_NAME)
#endif
#ifndef __SYCL_DATA_LESS_PROP_DEPRECATED_ALIAS
#define __SYCL_DATA_LESS_PROP_DEPRECATED_ALIAS(NS_QUALIFIER, PROP_NAME, ENUM_VAL, WARNING)
#endif

__SYCL_DATA_LESS_PROP(property, no_init, NoInit)
__SYCL_DATA_LESS_PROP_DEPRECATED_ALIAS(property, noinit, NoInit, __SYCL2020_DEPRECATED("spelling is now: no_init"))

#undef __SYCL_DATA_LESS_PROP
#undef __SYCL_MANUALLY_DEFINED_PROP
#undef __SYCL_DATA_LESS_PROP_DEPRECATED_ALIAS
