__SYCL_ASPECT(cpu, 1)
__SYCL_ASPECT(gpu, 2)
__SYCL_ASPECT(accelerator, 3)
__SYCL_ASPECT(custom, 4)
__SYCL_ASPECT(fp16, 5)
__SYCL_ASPECT(fp64, 6)
__SYCL_ASPECT(image, 9)
__SYCL_ASPECT(online_compiler, 10)
__SYCL_ASPECT(online_linker, 11)
__SYCL_ASPECT(queue_profiling, 12)
__SYCL_ASPECT(usm_device_allocations, 13)
__SYCL_ASPECT(usm_host_allocations, 14)
__SYCL_ASPECT(usm_shared_allocations, 15)
__SYCL_ASPECT(usm_system_allocations, 17)
__SYCL_ASPECT(ext_intel_pci_address, 18)
__SYCL_ASPECT(ext_intel_gpu_eu_count, 19)
__SYCL_ASPECT(ext_intel_gpu_eu_simd_width, 20)
__SYCL_ASPECT(ext_intel_gpu_slices, 21)
__SYCL_ASPECT(ext_intel_gpu_subslices_per_slice, 22)
__SYCL_ASPECT(ext_intel_gpu_eu_count_per_subslice, 23)
__SYCL_ASPECT(ext_intel_max_mem_bandwidth, 24)
__SYCL_ASPECT(usm_atomic_host_allocations, 25)
__SYCL_ASPECT(usm_atomic_shared_allocations, 26)
__SYCL_ASPECT(atomic64, 27)
__SYCL_ASPECT(ext_intel_device_info_uuid, 28)
__SYCL_ASPECT(ext_oneapi_srgb, 29)
__SYCL_ASPECT(ext_oneapi_native_assert, 30)
__SYCL_ASPECT(host_debuggable, 31)
__SYCL_ASPECT(ext_intel_gpu_hw_threads_per_eu, 32)
__SYCL_ASPECT(ext_oneapi_cuda_async_barrier, 33)
__SYCL_ASPECT(ext_intel_free_memory, 34)
__SYCL_ASPECT(ext_intel_device_id, 35)
__SYCL_ASPECT(ext_intel_memory_clock_rate, 36)
__SYCL_ASPECT(ext_intel_memory_bus_width, 37)
__SYCL_ASPECT(emulated, 38)
__SYCL_ASPECT(ext_intel_legacy_image, 39)
__SYCL_ASPECT(ext_oneapi_bindless_images, 40)
__SYCL_ASPECT(ext_oneapi_bindless_images_shared_usm, 41)
__SYCL_ASPECT(ext_oneapi_bindless_images_1d_usm, 42)
__SYCL_ASPECT(ext_oneapi_bindless_images_2d_usm, 43)
__SYCL_ASPECT(ext_oneapi_external_memory_import, 44)
__SYCL_ASPECT(ext_oneapi_external_semaphore_import, 45)
__SYCL_ASPECT(ext_oneapi_mipmap, 46)
__SYCL_ASPECT(ext_oneapi_mipmap_anisotropy, 47)
__SYCL_ASPECT(ext_oneapi_mipmap_level_reference, 48)
__SYCL_ASPECT(ext_intel_esimd, 49)
__SYCL_ASPECT(ext_oneapi_fragment, 50)
__SYCL_ASPECT(ext_oneapi_chunk, 51)
__SYCL_ASPECT(ext_oneapi_tangle, 52)
__SYCL_ASPECT(ext_intel_matrix, 53)
__SYCL_ASPECT(ext_oneapi_is_composite, 54)
__SYCL_ASPECT(ext_oneapi_is_component, 55)
__SYCL_ASPECT(ext_oneapi_graph, 56)
__SYCL_ASPECT(ext_oneapi_limited_graph, 57)
__SYCL_ASPECT(ext_oneapi_private_alloca, 58)
__SYCL_ASPECT(ext_oneapi_cubemap, 59)
__SYCL_ASPECT(ext_oneapi_cubemap_seamless_filtering, 60)
__SYCL_ASPECT(ext_oneapi_bindless_sampled_image_fetch_1d_usm, 61)
__SYCL_ASPECT(ext_oneapi_bindless_sampled_image_fetch_1d, 62)
__SYCL_ASPECT(ext_oneapi_bindless_sampled_image_fetch_2d_usm, 63)
__SYCL_ASPECT(ext_oneapi_bindless_sampled_image_fetch_2d, 64)
__SYCL_ASPECT(ext_oneapi_bindless_sampled_image_fetch_3d, 65)
__SYCL_ASPECT(ext_oneapi_queue_profiling_tag, 66)
__SYCL_ASPECT(ext_oneapi_virtual_mem, 67)
__SYCL_ASPECT(ext_oneapi_cuda_cluster_group, 68)
__SYCL_ASPECT(ext_oneapi_image_array, 69)
__SYCL_ASPECT(ext_oneapi_unique_addressing_per_dim, 70)
__SYCL_ASPECT(ext_oneapi_bindless_images_sample_1d_usm, 71)
__SYCL_ASPECT(ext_oneapi_bindless_images_sample_2d_usm, 72)
__SYCL_ASPECT(ext_oneapi_atomic16, 73)
__SYCL_ASPECT(ext_oneapi_virtual_functions, 74)
__SYCL_ASPECT(ext_intel_spill_memory_size, 75)
__SYCL_ASPECT(ext_oneapi_bindless_images_gather, 76)
__SYCL_ASPECT(ext_intel_current_clock_throttle_reasons, 77)
__SYCL_ASPECT(ext_intel_fan_speed, 78)
__SYCL_ASPECT(ext_intel_power_limits, 79)
__SYCL_ASPECT(ext_oneapi_async_memory_alloc, 80)
__SYCL_ASPECT(ext_intel_device_info_luid, 81)
__SYCL_ASPECT(ext_intel_device_info_node_mask, 82)
__SYCL_ASPECT(ext_oneapi_exportable_device_mem, 83)
__SYCL_ASPECT(ext_oneapi_clock_sub_group, 84)
__SYCL_ASPECT(ext_oneapi_clock_work_group, 85)
__SYCL_ASPECT(ext_oneapi_clock_device, 86)
__SYCL_ASPECT(ext_oneapi_is_integrated_gpu, 87)
__SYCL_ASPECT(ext_oneapi_ipc_memory, 88)
__SYCL_ASPECT(ext_oneapi_device_wait, 89)
