//==--------- graph.hpp --- SYCL graph extension ---------------------------==//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#pragma once

#include "graph/command_graph.hpp"
#include "graph/common.hpp"
#include "graph/dynamic.hpp"
#include "graph/executable_graph.hpp"
#include "graph/modifiable_graph.hpp"
#include "graph/node.hpp"
