//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef __SYCL_KHR_INCLUDES_HANDLER
#define __SYCL_KHR_INCLUDES_HANDLER

#include "version.hpp"

#include <sycl/handler.hpp>

#endif // __SYCL_KHR_INCLUDES_HANDLER
