
/*
 *
 * Copyright (C) 2024 Intel Corporation
 *
 * Part of the Unified-Runtime Project, under the Apache License v2.0 with LLVM
 * Exceptions.
 *
 * See LICENSE.TXT
 *
 * SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
 *
 * @file ur_api_funcs.def
 * @version v0.12-r0
 *
 */

// Auto-generated file, do not edit.

_UR_API(urAdapterGet)
_UR_API(urAdapterRelease)
_UR_API(urAdapterRetain)
_UR_API(urAdapterGetLastError)
_UR_API(urAdapterGetInfo)
_UR_API(urAdapterSetLoggerCallback)
_UR_API(urAdapterSetLoggerCallbackLevel)
_UR_API(urPlatformGet)
_UR_API(urPlatformGetInfo)
_UR_API(urPlatformGetNativeHandle)
_UR_API(urPlatformCreateWithNativeHandle)
_UR_API(urPlatformGetApiVersion)
_UR_API(urPlatformGetBackendOption)
_UR_API(urContextCreate)
_UR_API(urContextRetain)
_UR_API(urContextRelease)
_UR_API(urContextGetInfo)
_UR_API(urContextGetNativeHandle)
_UR_API(urContextCreateWithNativeHandle)
_UR_API(urContextSetExtendedDeleter)
_UR_API(urEventGetInfo)
_UR_API(urEventGetProfilingInfo)
_UR_API(urEventWait)
_UR_API(urEventRetain)
_UR_API(urEventRelease)
_UR_API(urEventGetNativeHandle)
_UR_API(urEventCreateWithNativeHandle)
_UR_API(urEventSetCallback)
_UR_API(urProgramCreateWithIL)
_UR_API(urProgramCreateWithBinary)
_UR_API(urProgramBuild)
_UR_API(urProgramCompile)
_UR_API(urProgramLink)
_UR_API(urProgramRetain)
_UR_API(urProgramRelease)
_UR_API(urProgramGetFunctionPointer)
_UR_API(urProgramGetGlobalVariablePointer)
_UR_API(urProgramGetInfo)
_UR_API(urProgramGetBuildInfo)
_UR_API(urProgramSetSpecializationConstants)
_UR_API(urProgramGetNativeHandle)
_UR_API(urProgramCreateWithNativeHandle)
_UR_API(urProgramDynamicLinkExp)
_UR_API(urProgramBuildExp)
_UR_API(urProgramCompileExp)
_UR_API(urProgramLinkExp)
_UR_API(urKernelCreate)
_UR_API(urKernelGetInfo)
_UR_API(urKernelGetGroupInfo)
_UR_API(urKernelGetSubGroupInfo)
_UR_API(urKernelRetain)
_UR_API(urKernelRelease)
_UR_API(urKernelGetNativeHandle)
_UR_API(urKernelCreateWithNativeHandle)
_UR_API(urKernelGetSuggestedLocalWorkSize)
_UR_API(urKernelSetArgValue)
_UR_API(urKernelSetArgLocal)
_UR_API(urKernelSetArgPointer)
_UR_API(urKernelSetExecInfo)
_UR_API(urKernelSetArgSampler)
_UR_API(urKernelSetArgMemObj)
_UR_API(urKernelSetSpecializationConstants)
_UR_API(urKernelSuggestMaxCooperativeGroupCount)
_UR_API(urQueueGetInfo)
_UR_API(urQueueCreate)
_UR_API(urQueueRetain)
_UR_API(urQueueRelease)
_UR_API(urQueueGetNativeHandle)
_UR_API(urQueueCreateWithNativeHandle)
_UR_API(urQueueFinish)
_UR_API(urQueueFlush)
_UR_API(urQueueBeginGraphCaptureExp)
_UR_API(urQueueBeginCaptureIntoGraphExp)
_UR_API(urQueueEndGraphCaptureExp)
_UR_API(urQueueIsGraphCaptureEnabledExp)
_UR_API(urSamplerCreate)
_UR_API(urSamplerRetain)
_UR_API(urSamplerRelease)
_UR_API(urSamplerGetInfo)
_UR_API(urSamplerGetNativeHandle)
_UR_API(urSamplerCreateWithNativeHandle)
_UR_API(urMemImageCreate)
_UR_API(urMemBufferCreate)
_UR_API(urMemRetain)
_UR_API(urMemRelease)
_UR_API(urMemBufferPartition)
_UR_API(urMemGetNativeHandle)
_UR_API(urMemBufferCreateWithNativeHandle)
_UR_API(urMemImageCreateWithNativeHandle)
_UR_API(urMemGetInfo)
_UR_API(urMemImageGetInfo)
_UR_API(urPhysicalMemCreate)
_UR_API(urPhysicalMemRetain)
_UR_API(urPhysicalMemRelease)
_UR_API(urPhysicalMemGetInfo)
_UR_API(urEnqueueKernelLaunch)
_UR_API(urEnqueueEventsWait)
_UR_API(urEnqueueEventsWaitWithBarrier)
_UR_API(urEnqueueMemBufferRead)
_UR_API(urEnqueueMemBufferWrite)
_UR_API(urEnqueueMemBufferReadRect)
_UR_API(urEnqueueMemBufferWriteRect)
_UR_API(urEnqueueMemBufferCopy)
_UR_API(urEnqueueMemBufferCopyRect)
_UR_API(urEnqueueMemBufferFill)
_UR_API(urEnqueueMemImageRead)
_UR_API(urEnqueueMemImageWrite)
_UR_API(urEnqueueMemImageCopy)
_UR_API(urEnqueueMemBufferMap)
_UR_API(urEnqueueMemUnmap)
_UR_API(urEnqueueUSMFill)
_UR_API(urEnqueueUSMMemcpy)
_UR_API(urEnqueueUSMPrefetch)
_UR_API(urEnqueueUSMAdvise)
_UR_API(urEnqueueUSMFill2D)
_UR_API(urEnqueueUSMMemcpy2D)
_UR_API(urEnqueueDeviceGlobalVariableWrite)
_UR_API(urEnqueueDeviceGlobalVariableRead)
_UR_API(urEnqueueReadHostPipe)
_UR_API(urEnqueueWriteHostPipe)
_UR_API(urEnqueueEventsWaitWithBarrierExt)
_UR_API(urEnqueueKernelLaunchWithArgsExp)
_UR_API(urEnqueueUSMDeviceAllocExp)
_UR_API(urEnqueueUSMSharedAllocExp)
_UR_API(urEnqueueUSMHostAllocExp)
_UR_API(urEnqueueUSMFreeExp)
_UR_API(urEnqueueCommandBufferExp)
_UR_API(urEnqueueTimestampRecordingExp)
_UR_API(urEnqueueHostTaskExp)
_UR_API(urEnqueueNativeCommandExp)
_UR_API(urEnqueueGraphExp)
_UR_API(urUSMHostAlloc)
_UR_API(urUSMDeviceAlloc)
_UR_API(urUSMSharedAlloc)
_UR_API(urUSMFree)
_UR_API(urUSMGetMemAllocInfo)
_UR_API(urUSMPoolCreate)
_UR_API(urUSMPoolRetain)
_UR_API(urUSMPoolRelease)
_UR_API(urUSMPoolGetInfo)
_UR_API(urUSMPoolCreateExp)
_UR_API(urUSMPoolDestroyExp)
_UR_API(urUSMPoolGetDefaultDevicePoolExp)
_UR_API(urUSMPoolGetInfoExp)
_UR_API(urUSMPoolSetInfoExp)
_UR_API(urUSMPoolSetDevicePoolExp)
_UR_API(urUSMPoolGetDevicePoolExp)
_UR_API(urUSMPoolTrimToExp)
_UR_API(urUSMPitchedAllocExp)
_UR_API(urUSMContextMemcpyExp)
_UR_API(urUSMImportExp)
_UR_API(urUSMReleaseExp)
_UR_API(urBindlessImagesUnsampledImageHandleDestroyExp)
_UR_API(urBindlessImagesSampledImageHandleDestroyExp)
_UR_API(urBindlessImagesImageAllocateExp)
_UR_API(urBindlessImagesImageFreeExp)
_UR_API(urBindlessImagesUnsampledImageCreateExp)
_UR_API(urBindlessImagesSampledImageCreateExp)
_UR_API(urBindlessImagesImageCopyExp)
_UR_API(urBindlessImagesImageGetInfoExp)
_UR_API(urBindlessImagesGetImageMemoryHandleTypeSupportExp)
_UR_API(urBindlessImagesGetImageUnsampledHandleSupportExp)
_UR_API(urBindlessImagesGetImageSampledHandleSupportExp)
_UR_API(urBindlessImagesMipmapGetLevelExp)
_UR_API(urBindlessImagesMipmapFreeExp)
_UR_API(urBindlessImagesImportExternalMemoryExp)
_UR_API(urBindlessImagesMapExternalArrayExp)
_UR_API(urBindlessImagesMapExternalLinearMemoryExp)
_UR_API(urBindlessImagesReleaseExternalMemoryExp)
_UR_API(urBindlessImagesFreeMappedLinearMemoryExp)
_UR_API(urBindlessImagesSupportsImportingHandleTypeExp)
_UR_API(urBindlessImagesImportExternalSemaphoreExp)
_UR_API(urBindlessImagesReleaseExternalSemaphoreExp)
_UR_API(urBindlessImagesWaitExternalSemaphoreExp)
_UR_API(urBindlessImagesSignalExternalSemaphoreExp)
_UR_API(urCommandBufferCreateExp)
_UR_API(urCommandBufferRetainExp)
_UR_API(urCommandBufferReleaseExp)
_UR_API(urCommandBufferFinalizeExp)
_UR_API(urCommandBufferAppendKernelLaunchExp)
_UR_API(urCommandBufferAppendUSMMemcpyExp)
_UR_API(urCommandBufferAppendUSMFillExp)
_UR_API(urCommandBufferAppendMemBufferCopyExp)
_UR_API(urCommandBufferAppendMemBufferWriteExp)
_UR_API(urCommandBufferAppendMemBufferReadExp)
_UR_API(urCommandBufferAppendMemBufferCopyRectExp)
_UR_API(urCommandBufferAppendMemBufferWriteRectExp)
_UR_API(urCommandBufferAppendMemBufferReadRectExp)
_UR_API(urCommandBufferAppendMemBufferFillExp)
_UR_API(urCommandBufferAppendUSMPrefetchExp)
_UR_API(urCommandBufferAppendUSMAdviseExp)
_UR_API(urCommandBufferAppendNativeCommandExp)
_UR_API(urCommandBufferUpdateKernelLaunchExp)
_UR_API(urCommandBufferUpdateSignalEventExp)
_UR_API(urCommandBufferUpdateWaitEventsExp)
_UR_API(urCommandBufferGetInfoExp)
_UR_API(urCommandBufferGetNativeHandleExp)
_UR_API(urGraphCreateExp)
_UR_API(urGraphInstantiateGraphExp)
_UR_API(urGraphDestroyExp)
_UR_API(urGraphExecutableGraphDestroyExp)
_UR_API(urGraphIsEmptyExp)
_UR_API(urGraphDumpContentsExp)
_UR_API(urIPCGetMemHandleExp)
_UR_API(urIPCPutMemHandleExp)
_UR_API(urIPCOpenMemHandleExp)
_UR_API(urIPCCloseMemHandleExp)
_UR_API(urMemoryExportAllocExportableMemoryExp)
_UR_API(urMemoryExportFreeExportableMemoryExp)
_UR_API(urMemoryExportExportMemoryHandleExp)
_UR_API(urUsmP2PEnablePeerAccessExp)
_UR_API(urUsmP2PDisablePeerAccessExp)
_UR_API(urUsmP2PPeerAccessGetInfoExp)
_UR_API(urVirtualMemGranularityGetInfo)
_UR_API(urVirtualMemReserve)
_UR_API(urVirtualMemFree)
_UR_API(urVirtualMemMap)
_UR_API(urVirtualMemUnmap)
_UR_API(urVirtualMemSetAccess)
_UR_API(urVirtualMemGetInfo)
_UR_API(urDeviceGet)
_UR_API(urDeviceGetInfo)
_UR_API(urDeviceRetain)
_UR_API(urDeviceRelease)
_UR_API(urDevicePartition)
_UR_API(urDeviceSelectBinary)
_UR_API(urDeviceGetNativeHandle)
_UR_API(urDeviceCreateWithNativeHandle)
_UR_API(urDeviceGetGlobalTimestamps)
_UR_API(urDeviceWaitExp)
_UR_API(urLoaderConfigCreate)
_UR_API(urLoaderConfigEnableLayer)
_UR_API(urLoaderConfigGetInfo)
_UR_API(urLoaderConfigRelease)
_UR_API(urLoaderConfigRetain)
_UR_API(urLoaderConfigSetCodeLocationCallback)
_UR_API(urLoaderConfigSetMockingEnabled)
_UR_API(urLoaderInit)
_UR_API(urLoaderTearDown)
