//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef __SYCL_KHR_INCLUDES_GROUPS
#define __SYCL_KHR_INCLUDES_GROUPS

#include "version.hpp"

#include <sycl/group.hpp>
#include <sycl/group_algorithm.hpp>
#include <sycl/group_barrier.hpp>
#include <sycl/sub_group.hpp>

#endif // __SYCL_KHR_INCLUDES_GROUPS
