// --*- c++ -*---
#ifndef __SYCL_DATA_LESS_PROP
#define __SYCL_DATA_LESS_PROP(NS_QUALIFIER, PROP_NAME, ENUM_VAL)
#endif
#ifndef __SYCL_MANUALLY_DEFINED_PROP
#define __SYCL_MANUALLY_DEFINED_PROP(NS_QUALIFIER, PROP_NAME)
#endif

__SYCL_DATA_LESS_PROP(property::buffer, use_host_ptr, BufferUseHostPtr)
__SYCL_DATA_LESS_PROP(ext::oneapi::property::buffer, use_pinned_host_memory, BufferUsePinnedHostMemory)

// Contains data field, defined explicitly.
__SYCL_MANUALLY_DEFINED_PROP(property::buffer, use_mutex)
__SYCL_MANUALLY_DEFINED_PROP(property::buffer, context_bound)

#undef __SYCL_DATA_LESS_PROP
#undef __SYCL_MANUALLY_DEFINED_PROP
