__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, global_work_size, sycl::range<3>, UR_KERNEL_GROUP_INFO_GLOBAL_WORK_SIZE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, work_group_size, size_t, UR_KERNEL_GROUP_INFO_WORK_GROUP_SIZE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, compile_work_group_size,
                  sycl::range<3>, UR_KERNEL_GROUP_INFO_COMPILE_WORK_GROUP_SIZE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific,
                  preferred_work_group_size_multiple, size_t, UR_KERNEL_GROUP_INFO_PREFERRED_WORK_GROUP_SIZE_MULTIPLE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, private_mem_size, size_t, UR_KERNEL_GROUP_INFO_PRIVATE_MEM_SIZE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, max_num_sub_groups, uint32_t, UR_KERNEL_SUB_GROUP_INFO_MAX_NUM_SUB_GROUPS)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, compile_num_sub_groups, uint32_t, UR_KERNEL_SUB_GROUP_INFO_COMPILE_NUM_SUB_GROUPS)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, max_sub_group_size, uint32_t, UR_KERNEL_SUB_GROUP_INFO_MAX_SUB_GROUP_SIZE)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, compile_sub_group_size, uint32_t,UR_KERNEL_SUB_GROUP_INFO_SUB_GROUP_SIZE_INTEL)
__SYCL_PARAM_TRAITS_SPEC(kernel_device_specific, ext_codeplay_num_regs, uint32_t, UR_KERNEL_INFO_NUM_REGS)
